/*
    Calculate Digits of Pi v0.00.00
    Copyright (C) 2007 Travis Evans
    This program contains modified code from www.codecodex.com
    http://www.codecodex.com/wiki/index.php?title=Digits_of_pi_calculation#C
    The CodeCodex main page grants permission to use code in other
    projects.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/* Program creation date:  2007/03/11
   Compilation command used:  tigcc -O3 -Wall pi.c
*/

#define USE_TI89
#define ENABLE_ERROR_RETURN
#define RETURN_VALUE

#include <alloc.h>
#include <args.h>
#include <error.h>
#include <estack.h>
#include <kbd.h>
#include <mem.h>
#include <statline.h>
#include <stdio.h>
#include <system.h>

void _main(void) {
    ESI argptr = top_estack;
    unsigned long arraySize, numDigits, bufferPos = 0;
    signed long *arr;
    char *returnStr, statmsg[65];
    
    /* Prevent AMS from detecting ON key on exit--this prevents a
       memory leak that apparently occurs due to pushing the
       return result to the expression stack without AMS ever using
       it when a Break Error occurs */
    OSDisableBreak();

    // Check argument
    if (GetArgType(argptr) == POSINT_TAG) {
        numDigits = GetIntArg(argptr);
    } else {
        ER_throw(ER_ARGUMENT);
    }
    if (numDigits < 4) numDigits = 4;   // numDigits must be at least 4
    // numDigits needs to be a multiple of four:
    while ((numDigits % 4) != 0) numDigits++;
    arraySize = (numDigits/4) * 14;
    
    // Allocate memory
    arr = (signed long *)calloc(arraySize+1, sizeof(signed long));
    returnStr = (unsigned char *)calloc(numDigits+2, sizeof(unsigned char));
    if (!(arr && returnStr)) {
        //Not enough memory; exit
        
        if (arr) free(arr);
        if (returnStr) free(returnStr);
        
        // I'm not sure if this is necessary, but just in case:
        while (GetArgType (top_estack) != END_TAG)  // Clean up arguments
            top_estack = next_expression_index (top_estack);
        top_estack--;

        // Exit with memory error
        ER_THROW(ER_MEMORY);
    }
    ST_helpMsg("Calculation started");
    
    // Perform the calculation
    #define SCALE 10000
    #define ARRINIT 2000
    signed long i, j;
    signed long carry = 0;
    for (i = 0; i <= arraySize; ++i)
            arr[i] = ARRINIT;
    for (i = arraySize; i > 0; i -= 14) {
            unsigned long sum = 0;
            for (j = i; j > 0; --j) {
                    sum = sum*j + SCALE*arr[j];
                    arr[j] = sum % (j*2-1);
                    sum /= (j*2-1);
            }
            // Write digits to buffer:
            sprintf(returnStr + bufferPos*sizeof(char), "%04lu",
                    carry + sum/SCALE);
            bufferPos += 4;
            // Display progress indicator:
            sprintf(statmsg, "%d%% complete; press a key to stop",
                (short)((100 * bufferPos) / numDigits));
            ST_helpMsg(statmsg);
            if (kbhit()) {
                ngetchx();
                returnStr[bufferPos+1] = 0;  // Terminate return string
                break;
            }
            carry = sum % SCALE;
    }
    
    // Return result:
    free(arr);
    ST_eraseHelp();
    memmove(returnStr + 2*sizeof(char), returnStr + 1*sizeof(char),
        numDigits-1);       // Make room for decimal point
    returnStr[1] = '.';     // Insert decimal point
    returnStr[numDigits+1] = 0;     // Make sure string is properly terminated
    while (GetArgType (top_estack) != END_TAG)  // Clean up arguments
        top_estack = next_expression_index (top_estack);
    top_estack--;
    push_zstr(returnStr);
    free(returnStr);
}
